#include "Arduino.h"
#include "SoftwareSerial.h"
#include "AltSoftSerial.h"
#include "DFRobotDFPlayerMini.h"
#include <LiquidCrystal_I2C.h>

int buttonNext = 4;
int buttonPause = 3;
int buttonPrevious = 2;
bool isPlaying = false;
uint8_t currentVolume = 1;
unsigned long t_debounce_1, t_debounce_2, t_debounce_3, t_debounce_4, t_debounce_5 = 0;
int index = 0;
int currentTrackNumber = 1;
char words[10];
char *songs[] = {
  "Humble-Kendrick Lamar",
  "Redbone-Childish Gambino",
  "The Effects of Climate Change-People Under The Stairs",
  "Starboy-The Weeknd",
  "Swimming Pools-Kendrick Lamar",
  "Break The Rules-Simon Viklund",
  "Virtual Insanity-Jamiroquai",
  "Lady Hear Me Tonight-Modjo",
  "Get Lucky-Daft Punk",
  "Broken Boy-Cage The Elephant",
  "Black Madonna-Cage The Elephant",
  "New Person, Same Old Mistakes-Tame Impala",
  "Come As You Are-Nirvana",
  "Lithium-Nirvana",
  "Can't Stop-Red Hot Chilli Peppers",
  "The Pretender-Foo Fighters",
  "Duality-Slipknot",
  "Psychosocial-Slipknot"
};

unsigned long track_times[] = {
  184000,
  327000,
  252000,
  273000,
  231000,
  254000,
  236000,
  221000,
  248000,
  164000,
  228000,
  364000,
  219000,
  257000,
  278000,
  270000,
  214000,
  302000
};

byte zero[] = {
  B00000,
  B00000,
  B00000,
  B00000,
  B00000,
  B00000,
  B00000
};

byte one[] = {
  B10000,
  B10000,
  B10000,
  B10000,
  B10000,
  B10000,
  B10000,
  B10000
};

byte two[] = {
  B11000,
  B11000,
  B11000,
  B11000,
  B11000,
  B11000,
  B11000,
  B11000
};

byte three[] = {
  B11100,
  B11100,
  B11100,
  B11100,
  B11100,
  B11100,
  B11100,
  B11100
};

byte four[] = {
  B11110,
  B11110,
  B11110,
  B11110,
  B11110,
  B11110,
  B11110,
  B11110
};

byte five[] = {
  B11111,
  B11111,
  B11111,
  B11111,
  B11111,
  B11111,
  B11111,
  B11111
};

unsigned long previousMillis = 0;
const long intervalLCD = 500; 
String topRowText;
int stringIndex = 0;
unsigned long startTime = 0;
unsigned long pausedTime = 0;
bool skipped = false;
bool paused = false;
bool resumed = false;
bool raisedVol = false;
bool decreasedVol = false;
bool playerInitial = true;
bool startPlaying = false;

SoftwareSerial mySoftwareSerial(10, 11); // RX, TX
AltSoftSerial bt;
DFRobotDFPlayerMini myDFPlayer;
LiquidCrystal_I2C lcd(0x27,16,2);
void printDetail(uint8_t type, int value);

void clearLCDLine(int line){
  lcd.setCursor(0,line);
  lcd.print("                ");
}

void handlePrev() {
  if (playerInitial) {
    startPlaying = true;
    return;
  }

  if (!isPlaying) {
    isPlaying = true;
    Serial.println(F("Player resumed!"));
    bt.println(F("Player resumed!"));
  }
  myDFPlayer.previous();
  if (currentTrackNumber == 1) {
    currentTrackNumber = 18;
  } else 
    currentTrackNumber--;
  topRowText = songs[currentTrackNumber -1];
  stringIndex = 0;
  startTime = millis();
  skipped = true;
  Serial.println(F("Skipped to the previous song!"));
  bt.println(F("Skipped to the previous song!"));
}

void handleNext() {
  if (playerInitial) {
    startPlaying = true;
    return;
  }

  if (!isPlaying) {
    isPlaying = true;
    Serial.println(F("Player resumed!"));
    bt.println(F("Player resumed!"));
  }
  myDFPlayer.next();
  if (currentTrackNumber == 18)
    currentTrackNumber = 1;
  else
    currentTrackNumber++;
  topRowText = songs[currentTrackNumber - 1];
  stringIndex = 0;
  startTime = millis();
  skipped = true;
  Serial.println(F("Skipped to the next song!"));
  bt.println(F("Skipped to the next song!"));
}

void handleVolumeUp() {
  if (playerInitial) {
    startPlaying = true;
    return;
  }

  if (currentVolume < 30) {
    myDFPlayer.volumeUp();
    currentVolume++;
    Serial.print(F("Volume increased! Volume Level: "));
    Serial.println(currentVolume);
    bt.print(F("Volume increased! Volume Level: "));
    bt.println(currentVolume);
  } else {
    Serial.println("MAX VOLUME!");
    bt.println("MAX VOLUME!");
  }
  raisedVol = true;
}

void handleVolumeDown() {
  if (playerInitial) {
    startPlaying = true;
    return;
  }

  if (currentVolume > 0) {
    myDFPlayer.volumeDown();
    currentVolume--;
    Serial.print(F("Volume decreased! Volume Level: "));
    Serial.println(currentVolume);
    bt.print(F("Volume decreased! Volume Level: "));
    bt.println(currentVolume);
  } else {
    Serial.println("Player muted!");
    bt.println("Player muted!");
  }
  decreasedVol = true;
}

void handlePause() {
  if (playerInitial) {
    startPlaying = true;
    return;
  }

  if (isPlaying) {
    myDFPlayer.pause();
    Serial.println(F("Player paused!"));
    bt.println(F("Player paused!"));
    isPlaying = !isPlaying;
    pausedTime = millis() - startTime;
    startTime = 0;
    paused = true;
    resumed = false;
  } 
}

void handleResume() {
  if (playerInitial) {
    startPlaying = true;
    return;
  }
  
  if (!isPlaying) {
    myDFPlayer.start();
    Serial.println(F("Player resumed!"));
    bt.println(F("Player resumed!"));
    isPlaying = !isPlaying;
    startTime = millis() - pausedTime;
    paused = false;
    resumed = true;
  }
}

void previous() {
  unsigned long t_int = millis();

  if (t_int - t_debounce_1 > 300) {
    t_debounce_1 = t_int;
    handlePrev();
  }
}

void pause() {
  unsigned long t_int = millis();

  if (t_int - t_debounce_2 > 300) {
    t_debounce_2 = t_int;
    if (isPlaying) {
      handlePause();
    } else {
      handleResume();
    }
  }
}

ISR (PCINT2_vect) {
  unsigned long t_int = millis();

  if ((PIND & (1 << PD4)) == 0) {
    if (t_int - t_debounce_3 > 300) {
      t_debounce_3 = t_int;
      handleNext();
    }
  } else if ((PIND & (1 << PD5)) == 0) {
    if (t_int - t_debounce_4 > 300) {
      t_debounce_4 = t_int;
      handleVolumeDown();
    }
  } else if ((PIND & (1 << PD6)) == 0) {
    if (t_int - t_debounce_5 > 300) {
      t_debounce_5 = t_int;
      handleVolumeUp();
    }
  }
  
}

void scrollInFromRight (int line, char str1[]) {
  for (int letter = 0; letter <= strlen(str1); letter++)
  {
    lcd.setCursor(0, 0);
    for (int i = letter; i <= letter + 15; i++) {
      if (i >= strlen(str1))
        lcd.print(" ");
      else
        lcd.print(str1[i]);
    }
    delay(500);
  }
}

void updateProgressBar(unsigned long count, unsigned long totalCount, int lineToPrintOn) {
  double factor = totalCount / 80.0;
  int percent = (count + 1) / factor;
  int number = percent / 5;
  int remainder = percent % 5;

  if (number > 0) {
    for (int i = 0; i < number; i++) {
      lcd.setCursor(i, lineToPrintOn);
      lcd.write(5);
    }
  }

  lcd.setCursor(number, lineToPrintOn);

  if (remainder > 0 && number < 16) {
    lcd.write(remainder);
  } else {
    lcd.write(' ');
  }
}

void initialState() {
  lcd.setCursor(0, 0);
  lcd.print("Player online!");
  lcd.setCursor(0, 1);
  lcd.print("Press any button");
  if (bt.available()) {
    words[index] = bt.read();

    if (words[index] == '\n') {
      words[index] = '\0';
      if (strncmp(words, "next", 4) == 0) {
        startPlaying = true;
      } else if (strncmp(words, "prev", 4) == 0) {
        startPlaying = true;
      } else if (strncmp(words, "pause", 5) == 0) {
        startPlaying = true;
      } else if (strncmp(words, "resume", 6) == 0) {
        startPlaying = true;
      } else if (strncmp(words, "volume+", 7) == 0) {
        startPlaying = true;
      } else if (strncmp(words, "volume-", 7) == 0) {
        startPlaying = true;
      } else if (strncmp(words, "start", 5) == 0) {
        startPlaying = true;
      }
      for (int i = 0; i < 10; i++) {
        words[i] = '\0';
      }
      index = 0;
    } else {
      index++;
    }

    if (index > 9) {
      Serial.println("Illegal word!");
      bt.println("Illegal word!");
      index = 0;
      for (int i = 0; i < 10; i++) {
        words[i] = '\0';
      }
    }
  }
}


void setup() {
  cli();
  // input
  DDRD &= ~(1 << PD2) & ~(1 << PD3) & ~(1 << PD4) & ~(1 << PD5) & ~(1 << PD6);
  // input pullup
  PORTD |= (1 << PD2) | (1 << PD3) | (1 << PD4) | (1 << PD5) | (1 << PD6);

  attachInterrupt(digitalPinToInterrupt(2), previous, FALLING);
  attachInterrupt(digitalPinToInterrupt(3), pause, FALLING);
  PCICR |= B00000100;
  PCMSK2 |= B01110000;
  sei();
  
  bt.begin(9600);
  mySoftwareSerial.begin(9600);
  Serial.begin(115200);
  
  Serial.println();
  Serial.println(F("Initializing DFPlayer..."));
  
  if (!myDFPlayer.begin(mySoftwareSerial)) {
    Serial.println(F("Unable to begin:"));
    Serial.println(F("1.Please recheck the connection!"));
    Serial.println(F("2.Please insert the SD card!"));
    while(true){
      delay(0);
    }
  }
  
  Serial.println(F("DFPlayer Mini online."));

  // Set LCD
  lcd.init();
  lcd.init();
  lcd.clear();         
  lcd.backlight();
  lcd.createChar(0, zero);
  lcd.createChar(1, one);
  lcd.createChar(2, two);
  lcd.createChar(3, three);
  lcd.createChar(4, four);
  lcd.createChar(5, five);
}

void loop() { 
  if (playerInitial) {
    initialState();
    if (startPlaying) {
      Serial.println(F("Player Started."));
      bt.println(F("Player Started."));
      myDFPlayer.volume(currentVolume);
      myDFPlayer.play(currentTrackNumber);
      isPlaying = true;
      topRowText = songs[currentTrackNumber - 1];
      playerInitial = false;
      clearLCDLine(1);
      startTime = millis();
    }
  } else {
    if (myDFPlayer.available()) {
      printDetail(myDFPlayer.readType(), myDFPlayer.read());
    }

    unsigned long currentMillis = millis();
    if (currentMillis - previousMillis >= intervalLCD) {
      lcd.setCursor(0, 0);
      lcd.print(topRowText.substring(stringIndex, stringIndex+16));
      lcd.print(" ");
      previousMillis = currentMillis; 
      stringIndex++;
      if (stringIndex >= topRowText.length()) {
        stringIndex = 0;
      }
    }

    if (skipped) {
      clearLCDLine(1);
      skipped = false;
      lcd.setCursor(0, 1);
      lcd.print("Skipping...");
      delay(300);
      clearLCDLine(0);
      clearLCDLine(1);
    }

    if (paused) {
      clearLCDLine(1);
      paused = false;
    }

    if (resumed) {
      clearLCDLine(1);
      resumed = false;
      lcd.setCursor(0, 1);
      lcd.print("Resuming...");
      delay(300);
      clearLCDLine(1);
    }

    if (raisedVol) {
      clearLCDLine(1);
      raisedVol = false;
      lcd.setCursor(0, 1);
      if (currentVolume < 30) {
        lcd.print("Vol++   Vol = ");
        lcd.setCursor(14, 1);
        lcd.print(currentVolume);
        delay(300);
        clearLCDLine(1);
      } else {
        lcd.print("MAX VOLUME");
        delay(300);
        clearLCDLine(1);
      }
    }

    if (decreasedVol) {
      clearLCDLine(1);
      decreasedVol = false;
      lcd.setCursor(0, 1);
      if (currentVolume >= 0) {
        lcd.print("Vol--   Vol = ");
        lcd.setCursor(14, 1);
        lcd.print(currentVolume);
        delay(300);
        clearLCDLine(1);
      } else {
        lcd.print("MIN VOLUME");
        delay(300);
        clearLCDLine(1);
      }
    }

    unsigned long elapsedTime = 0;
    if (startTime != 0) {
      elapsedTime = millis() - startTime;
      updateProgressBar(elapsedTime, track_times[currentTrackNumber - 1] - 2000, 1);
    } else {
      lcd.setCursor(0, 1);
      lcd.print("PAUSED");
    }

    if (bt.available()) {
      words[index] = bt.read();

      if (words[index] == '\n') {
        words[index] = '\0';
        if (strncmp(words, "next", 4) == 0) {
          handleNext();
        } else if (strncmp(words, "prev", 4) == 0) {
          handlePrev();
        } else if (strncmp(words, "pause", 5) == 0) {
          handlePause();
        } else if (strncmp(words, "resume", 6) == 0) {
          handleResume();
        } else if (strncmp(words, "volume+", 7) == 0) {
          handleVolumeUp();
        } else if (strncmp(words, "volume-", 7) == 0) {
          handleVolumeDown();
        }
        for (int i = 0; i < 10; i++) {
          words[i] = '\0';
        }
        index = 0;
      } else {
        index++;
      }

      if (index > 9) {
        Serial.println("Illegal word!");
        bt.println("Illegal word!");
        index = 0;
        for (int i = 0; i < 10; i++) {
          words[i] = '\0';
        }
      }
    }
  }
}

void printDetail(uint8_t type, int value) {
  switch (type) {
    case TimeOut:
      Serial.println(F("Time Out!"));
      break;
    case WrongStack:
      Serial.println(F("Stack Wrong!"));
      break;
    case DFPlayerCardInserted:
      Serial.println(F("Card Inserted!"));
      break;
    case DFPlayerCardRemoved:
      Serial.println(F("Card Removed!"));
      break;
    case DFPlayerCardOnline:
      Serial.println(F("Card Online!"));
      break;
    case DFPlayerUSBInserted:
      Serial.println("USB Inserted!");
      break;
    case DFPlayerUSBRemoved:
      Serial.println("USB Removed!");
      break;
    case DFPlayerPlayFinished:
      Serial.println(F(" Play Finished!"));
      myDFPlayer.next();
      currentTrackNumber++;
      topRowText = songs[currentTrackNumber - 1];
      stringIndex = 0;
      startTime = millis();
      skipped = true;
      break;
    case DFPlayerError:
      Serial.print(F("DFPlayerError:"));
      switch (value) {
        case Busy:
          Serial.println(F("Card not found"));
          break;
        case Sleeping:
          Serial.println(F("Sleeping"));
          break;
        case SerialWrongStack:
          Serial.println(F("Get Wrong Stack"));
          break;
        case CheckSumNotMatch:
          Serial.println(F("Check Sum Not Match"));
          break;
        case FileIndexOut:
          Serial.println(F("File Index Out of Bound"));
          break;
        case FileMismatch:
          Serial.println(F("Cannot Find File"));
          break;
        case Advertise:
          Serial.println(F("In Advertise"));
          break;
        default:
          break;
      }
      break;
    default:
      break;
  }
}
